/*
 * Copyright (c) 2010-2011 Hiroki Sato.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $hrs: linux_adobe/linux_adobe.c,v 1.3 2011/05/29 15:18:59 hrs Exp $
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/libkern.h>

#include <sys/proc.h>
#include <sys/syscallsubr.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/sched.h>

#ifdef COMPAT_LINUX32
#include <machine/../linux32/linux.h>
#include <machine/../linux32/linux32_proto.h>
#else
#include <machine/../linux/linux.h>
#include <machine/../linux/linux_proto.h>
#endif

#define LINUX_ADOBE_PROCNAME	"acroread"

extern struct sysent linux_sysent[];
#define SYS_linux_sched_setscheduler	(156)

static int
linux_adobe_linux_sched_setscheduler(struct thread *td,
	struct linux_sched_setscheduler_args *args)
{
	int error = 0;
	struct sched_setscheduler_args bsd;

	/* Fallback immediately for non acroread processes */
	if (fnmatch(LINUX_ADOBE_PROCNAME, td->td_proc->p_comm, FNM_NOESCAPE)
	    == FNM_NOMATCH)
		return (linux_sched_setscheduler(td, args));
#ifdef DEBUG
	if (ldebug(sched_setscheduler))
		printf(ARGS(sched_setscheduler, "%d, %d, %p"),
		    args->pid, args->policy, (const void *)args->param);
#endif
	switch (args->policy) {
	case LINUX_SCHED_OTHER:
		bsd.policy = SCHED_OTHER;
		break;
	case LINUX_SCHED_FIFO:
		bsd.policy = SCHED_FIFO;
		break;
	case LINUX_SCHED_RR:
		bsd.policy = SCHED_RR;
		break;
	default:
		return EINVAL;
	}

	bsd.pid = args->pid;
	bsd.param = (struct sched_param *)args->param;
	error = sched_setscheduler(td, &bsd);
	/*
	 * Adobe Reader can fail when a non-root user invoked this syscall.
	 * Ignore the error condition if the process's uid != 0.
	 */
	if (error &&
	    td != NULL &&
	    td->td_proc != NULL &&
	    td->td_proc->p_ucred->cr_ruid != 0)
		error = 0;

	return (error);
}

static int
linux_adobe_load(struct module *module, int cmd, void *arg)
{
	int error = 0;

	switch (cmd) {
	case MOD_LOAD:
		linux_sysent[SYS_linux_sched_setscheduler].sy_call =
			(sy_call_t *)linux_adobe_linux_sched_setscheduler;
		break;
	case MOD_UNLOAD:
		linux_sysent[SYS_linux_sched_setscheduler].sy_call =
			(sy_call_t *)linux_sched_setscheduler;
		break;
	default:
		error = EOPNOTSUPP;
		break;
	}

	return (error);
}

static moduledata_t linux_adobe_moddata = {
	"linux_adobe",
	linux_adobe_load,
	NULL
};
DECLARE_MODULE(linux_adobe, linux_adobe_moddata, SI_SUB_DRIVERS, SI_ORDER_MIDDLE);
MODULE_VERSION(linux_adobe, 1);
MODULE_DEPEND(linux_adobe, linux, 1, 1, 1);
